# Hello everyone, my name is Renzhai
# This API provides more customized extensions to the official API
# If you have any questions, please feel free to leave a message on http://renzhi.net
# Corresponding UE plugin address :https://renzhai.net/980.html

from turtle import Turtle
from fastapi import FastAPI, Request
from transformers import AutoTokenizer, AutoModel
import uvicorn, json, datetime
import torch
from sse_starlette.sse import ServerSentEvent, EventSourceResponse

import argparse

DEVICE = "cuda"
DEVICE_ID = "0"
CUDA_DEVICE = f"{DEVICE}:{DEVICE_ID}" if DEVICE_ID else DEVICE


def torch_gc():
    if torch.cuda.is_available():
        with torch.cuda.device(CUDA_DEVICE):
            torch.cuda.empty_cache()
            torch.cuda.ipc_collect()

async def chat_generator(prompt,history):
    global model, tokenizer
    response =""
    current_length = 0
    for response, history in model.stream_chat(tokenizer, prompt, history):
        if len(response) == current_length:
            continue
        new_text = response[current_length:]
        current_length = len(response)
        nest_answer = {
        "response": new_text,
        "status": 200,
        "complete":False
        }
        nest_answer_json = json.dumps(nest_answer, ensure_ascii=False)
        #print(nest_answer)
        yield ServerSentEvent(nest_answer_json)
    now = datetime.datetime.now()
    time = now.strftime("%Y-%m-%d %H:%M:%S")
    answer = {
    "response": response,
    "history": history,
    "status": 200,
    "time": time,
    "complete":True
    }
    nest_answer_json = json.dumps(answer, ensure_ascii=False)
    yield ServerSentEvent(nest_answer_json)
    torch_gc()

app = FastAPI()


@app.post("/")
async def create_item(request: Request):
    global model, tokenizer
    json_post_raw = await request.json()
    json_post = json.dumps(json_post_raw)
    json_post_list = json.loads(json_post)

    prompt = json_post_list.get('prompt')
    history = json_post_list.get('history')
    max_length = json_post_list.get('max_length')
    top_p = json_post_list.get('top_p')
    temperature = json_post_list.get('temperature')
    bstream = json_post_list.get('stream')
    if bstream:
       return EventSourceResponse(chat_generator(prompt,history),media_type="text/event-stream")

    response, history = model.chat(tokenizer,
                                   prompt,
                                   history=history,
                                   max_length=max_length if max_length else 2048,
                                   top_p=top_p if top_p else 0.7,
                                   temperature=temperature if temperature else 0.95)
    now = datetime.datetime.now()
    time = now.strftime("%Y-%m-%d %H:%M:%S")
    answer = {
        "response": response,
        "history": history,
        "status": 200,
        "complete":True,
        "time": time
    }
    #log = "[" + time + "] " + '", prompt:"' + prompt + '", response:"' + repr(response) + '"'
    #print(log)
    torch_gc()
    return answer


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='chatglm argparse')
    parser.add_argument('--model_path','-mp',type=str, default = ".\\model",required=True,help="Path of the model.")
    parser.add_argument('--ip','-ip',type=str, default = "0.0.0.0",required=False,help="Local IP")
    parser.add_argument('--port','-port',type=int, default = 8540,required=False,help="The address of the port")
    parser.add_argument('--workers','-wk',type=int, default = 1,required=False,help="How many threads are enabled defaults to 1.")

    args = parser.parse_args()

    tokenizer = AutoTokenizer.from_pretrained(args.model_path, trust_remote_code=True)
    model = AutoModel.from_pretrained(args.model_path, trust_remote_code=True).cuda()

    # 如果你的torch类型是 float16 那就开启这个，把上面的关闭
    #tokenizer = AutoTokenizer.from_pretrained(args.model_path, trust_remote_code=True,torch_dtype=torch.float16)
    #model = AutoModel.from_pretrained(args.model_path, trust_remote_code=True,torch_dtype=torch.float16).cuda()
       
    # 多显卡支持，使用下面三行代替上面两行，将num_gpus改为你实际的显卡数量
    #tokenizer = AutoTokenizer.from_pretrained(args.model_path, trust_remote_code=True)
    #model = load_model_on_gpus(args.model_path, num_gpus=args.GPU_num)

    model.eval()
    uvicorn.run(app, host=args.ip, port=args.port, workers=args.workers)
